#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""model module."""
import datetime
import logging
import os
from difflib import Differ
from urllib.parse import urljoin, urlparse

import structlog
from bs4 import BeautifulSoup
from peewee import (
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKeyField,
    IntegerField,
    Model,
    SqliteDatabase,
    TextField
)
from PIL import Image

from iqdb_tagger.sha256 import sha256_checksum
from iqdb_tagger.utils import default_db_path

db = SqliteDatabase(None)
log = structlog.getLogger()


class BaseModel(Model):
    """base model."""

    class Meta:
        """meta."""

        database = db


class Program(BaseModel):
    """program model."""

    version = IntegerField()


class Tag(BaseModel):
    """Tag model."""

    name = CharField()
    namespace = CharField(null=True)

    @property
    def full_name(self):
        """Get full name."""
        if self.namespace:
            return self.namespace + ':' + self.name
        return self.name


class Match(BaseModel):
    """Match model."""

    RATING_UNKNOWN = 0
    RATING_SAFE = 1
    RATING_ERO = 2
    RATING_EXPLICIT = 3

    RATING_CHOICES = (
        (RATING_UNKNOWN, 'Unknown'),
        (RATING_SAFE, 'Safe'),
        (RATING_ERO, 'Ero'),
        (RATING_EXPLICIT, 'Explicit'),
    )
    href = CharField(unique=True)
    thumb = CharField()
    rating = CharField()
    img_alt = TextField(null=True)
    width = IntegerField(null=True)
    height = IntegerField(null=True)

    @property
    def iqdb_thumb(self):
        """Get iqdb thumb url."""
        return urljoin('https://iqdb.org', self.thumb)

    @property
    def size(self):
        """Get size string."""
        if self.width and self.height:
            return '{}x{}'.format(self.width, self.height)

    @property
    def link(self):
        """Get href link."""
        return urljoin('https://', self.href)

    @property
    def link_netloc(self):
        """Get readable netloc."""
        netloc = urlparse(self.link).netloc
        if netloc.startswith('www.'):
            netloc = netloc.split('www.', 1)[1]
        endings = ['.net', '.com', '.us']
        for ending in endings:
            if netloc.endswith(ending):
                netloc = netloc.split(ending, 1)[0]
        return netloc

    @property
    def tags_from_img_alt(self):
        """Get readable tag from image alt."""
        result = []
        img_alt = self.img_alt[0]
        non_tags_txt = img_alt.split('Tags:')[0]
        tags_txt = img_alt.split('Tags:')[1]
        result.extend(tags_txt.split(' '))
        non_tags_txt.split('Score:')
        result.append(non_tags_txt.split('Score:')[0])
        result.append('Score:' + non_tags_txt.split('Score:')[1])
        result = [x.strip() for x in result if x]
        return result


class MatchTagRelationship(BaseModel):
    """match tag relationship."""

    match = ForeignKeyField(Match)
    tag = ForeignKeyField(Tag)


class ImageModel(BaseModel):
    """Image model."""

    checksum = CharField(unique=True)
    width = IntegerField()
    height = IntegerField()
    path = CharField(null=True)

    @property
    def size(self):
        """Get size string."""
        return '{}x{}'.format(self.width, self.height)

    @property
    def path_basename(self):
        """Get path basename."""
        return os.path.basename(self.path)

    @staticmethod
    def get_or_create_from_path(img_path):
        """Get or crate from path."""
        checksum = sha256_checksum(img_path)
        img = Image.open(img_path)
        width, height = img.size
        img, created = ImageModel.get_or_create(
            checksum=checksum, defaults={
                'width': width, 'height': height, 'path': img_path,
            }
        )
        return img, created

    def __str__(self):
        """Get string repr."""
        return '{}, checksum:{}..., size:{}x{} path:{}'.format(
            super().__str__(), self.checksum[:5],
            self.width, self.height, self.path
        )


class ImageMatchRelationship(BaseModel):
    """Image and match result."""

    image = ForeignKeyField(ImageModel)
    match_result = ForeignKeyField(Match)  # NOQA


class ImageMatch(BaseModel):
    """Image match."""

    STATUS_UNKNOWN = 0
    STATUS_BEST_MATCH = 1
    STATUS_POSSIBLE_MATCH = 2
    STATUS_OTHER = 3
    STATUS_CHOICES = (
        (STATUS_UNKNOWN, 'Unknown'),
        (STATUS_BEST_MATCH, 'Best match'),
        (STATUS_POSSIBLE_MATCH, 'Possible match'),
        (STATUS_OTHER, 'Other'),
    )
    SP_IQDB = 0
    SP_DANBOORU = 1
    SP_E621 = 2
    SP_CHOICES = (
        (SP_IQDB, 'iqdb'),
        (SP_DANBOORU, 'danbooru'),
        (SP_E621, 'e621'),
    )
    match = ForeignKeyField(ImageMatchRelationship)
    similarity = IntegerField()
    status = IntegerField(choices=STATUS_CHOICES)
    search_place = IntegerField(choices=SP_CHOICES)
    created_date = DateTimeField(default=datetime.datetime.now)
    force_gray = BooleanField(default=False)

    @staticmethod
    def _get_status_from_header_tag(header_tag):
        """Get status from header tag."""
        if hasattr(header_tag, 'text'):
            header_text = header_tag.text
            if header_text in ('Your image', 'No relevant matches'):
                return None
            best_match_text = \
                ('Best match', 'Additional match', 'Probable match:')
            if header_text == 'Possible match':
                status = ImageMatch.STATUS_POSSIBLE_MATCH
            elif header_text in best_match_text:
                status = ImageMatch.STATUS_BEST_MATCH
            elif header_text == 'Improbable match:':
                status = ImageMatch.STATUS_OTHER
            else:
                log.debug('header text', v=header_text)
                status = ImageMatch.STATUS_OTHER
        else:
            status = ImageMatch.STATUS_OTHER
        return status

    @staticmethod
    def parse_table(table):
        """Parse table."""
        header_tag = table.select_one('th')
        status = ImageMatch._get_status_from_header_tag(header_tag)
        if status is None:
            return None
        td_tags = table.select('td')
        assert '% similarity' in td_tags[-1].text, "similarity was not found in " + header_tag.text
        size_and_rating_text = td_tags[-2].text
        rating = Match.RATING_UNKNOWN
        for item in Match.RATING_CHOICES:
            if '[{}]'.format(item[1]) in size_and_rating_text:
                rating = item[0]
        size = size_and_rating_text.strip().split(' ', 1)[0].split('×')
        if len(size) == 1 and '×' not in size_and_rating_text:
            size = (None, None)
        else:
            size = (int(size[0]), int(size[1]))
        img_tag = table.select_one('img')
        img_alt = img_tag.attrs.get('alt')
        img_title = img_tag.attrs.get('title')
        if img_alt == '[IMG]' and img_title is None:
            img_alt = None
        if img_alt != img_title:
            d = Differ()
            diff_text = '\n'.join(d.compare(img_alt, img_title))
            log.warning(
                'title and alt attribute of img tag is different.\n{}'.format(
                    diff_text
                )
            )
        return {
            # match
            'status': status,
            'similarity': td_tags[-1].text.split('% similarity', 1)[0],
            # match result
            'href': table.select_one('a').attrs.get('href', None),
            'thumb': table.select_one('img').attrs.get('src', None),
            'rating': rating,
            'size': size,
            'img_alt': img_alt,
        }

    @staticmethod
    def parse_page(page):
        """Parse page."""
        if isinstance(page, str):
            page = BeautifulSoup(page, 'lxml')
        elif not isinstance(page, BeautifulSoup):
            if not os.path.isfile(page):
                raise ValueError('File not Exist: {}'.format(page))
            with open(page) as f:
                soup = BeautifulSoup(f.read(), 'lxml')
            page = soup

        tables = page.select('.pages table')
        for table in tables:
            res = ImageMatch.parse_table(table)
            if not res:
                continue
            a_tags = table.select('a')
            assert len(a_tags) < 3, "Unexpected html received at parse_page. Malformed link"
            if len(a_tags) == 2:
                additional_res = res
                additional_res['href'] = \
                    a_tags[1].attrs.get('href', None)
                yield additional_res
            yield res

    @staticmethod
    def get_or_create_from_page(page, image, place=None, force_gray=False):
        """Get or create from page result."""
        if place is None:
            place = ImageMatch.SP_IQDB
        items = ImageMatch.parse_page(page)
        for item in items:
            match_result, _ = Match.get_or_create(
                href=item['href'], defaults={
                    'thumb': item['thumb'],
                    'rating': item['rating'],
                    'img_alt': item['img_alt'],
                    'width': item['size'][0],
                    'height': item['size'][1],
                }
            )
            imr, _ = ImageMatchRelationship.get_or_create(
                image=image,
                match_result=match_result,
            )
            yield ImageMatch.get_or_create(
                match=imr,
                search_place=place,
                force_gray=force_gray,
                defaults={
                    'status': item['status'],
                    'similarity': item['similarity'],
                }
            )

    @property
    def status_verbose(self):
        """Get verbose status."""
        return dict(ImageMatch.STATUS_CHOICES)[self.status]

    @property
    def search_place_verbose(self):
        """Get verbose search place."""
        return dict(ImageMatch.SP_CHOICES)[self.search_place]


class ThumbnailRelationship(BaseModel):
    """Thumbnail tag relationship."""

    original = ForeignKeyField(ImageModel, related_name='thumbnails')  # NOQA
    thumbnail = ForeignKeyField(ImageModel)

    @staticmethod
    def get_or_create_from_image(
            image, size, thumb_folder=None, thumb_path=None, img_path=None):
        """Get or create from image."""
        thumbnails = [
            x for x in image.thumbnails
            if x.thumbnail.width == size[0] and x.thumbnail.height == size[1]
        ]
        if thumbnails:
            assert len(thumbnails) == 1, "There was not one thumbnail for the result"
            return thumbnails[0], False
        if thumb_path is None:
            thumb_path = '{}-{}-{}.jpg'.format(image.checksum, size[0], size[1])
            if thumb_folder:
                thumb_path = os.path.join(thumb_folder, thumb_path)
        if not os.path.isfile(thumb_path) or os.path.getsize(thumb_path) == 0:
            im = Image.open(image.path) if img_path is None else Image.open(img_path)
            im.thumbnail(size, Image.ANTIALIAS)
            try:
                im.save(thumb_path, 'JPEG')
            except OSError as e:
                valid_err = [
                    'cannot write mode RGBA as JPEG',
                    'cannot write mode P as JPEG',
                    'cannot write mode LA as JPEG',
                ]
                err_str = str(e)
                if err_str in valid_err:
                    log.debug('Converting to JPEG for error fix', err=err_str)
                    im = im.convert('RGB')
                    im.save(thumb_path, 'JPEG')
                else:
                    raise e
        thumb, _ = ImageModel.get_or_create_from_path(thumb_path)
        return ThumbnailRelationship.get_or_create(
            original=image, thumbnail=thumb)


def init_db(db_path=None, version=1):
    """Init db."""
    if db_path is None:
        db_path = default_db_path
    db.init(db_path)
    if not os.path.isfile(db_path):
        model_list = [
            ImageMatch,
            ImageMatchRelationship,
            ImageModel,
            Match,
            MatchTagRelationship,
            Program,
            Tag,
            ThumbnailRelationship
        ]
        db.create_tables(model_list)
        version = Program(version=version)
        version.save()
    else:
        logging.debug('db already existed.')



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""main module."""
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse
import logging
import os
import shutil

import cfscrape
import click
import mechanicalsoup
import requests
import structlog

from iqdb_tagger import models
from iqdb_tagger.__init__ import db_version
from iqdb_tagger.custom_parser import get_tags as get_tags_from_parser
from iqdb_tagger.utils import default_db_path, thumb_folder, user_data_dir

db = '~/images/! tagged'
DEFAULT_SIZE = 150, 150
DEFAULT_PLACE = 'iqdb'
minsim = 75
services = ['1', '2', '3', '4', '5', '6', '10', '11']
forcegray = False
log = structlog.getLogger()
iqdb_url_dict = {
    'iqdb': (
        'http://iqdb.org', models.ImageMatch.SP_IQDB),
    'danbooru': (
        'http://danbooru.iqdb.org', models.ImageMatch.SP_DANBOORU
    ),
    'e621': (
        'http://iqdb.harry.lu', models.ImageMatch.SP_E621),
}


def get_page_result(image, url, browser=None, use_requests=False):
    """Get page result.

    Args:
        image: Image path to be uploaded.
    Returns:
        HTML page from the result.
    """
    # compatibility
    br = browser

    if br is None:
        br = mechanicalsoup.StatefulBrowser(soup_config={'features': 'lxml'})
        br.raise_on_404 = True

    if use_requests:
        files = {'file': open(image, 'rb')}
        resp = requests.post(url, files=files, timeout=10)
        return resp.text
    br.open(url)
    html_form = br.select_form('form')
    html_form.input({'file': image})
    br.submit_selected()
    # if ok, will output: <Response [200]>
    return br.get_current_page()


def get_posted_image(
        img_path, resize=False, size=None, output_thumb_folder=None, thumb_path=None):
    """Get posted image."""
    if output_thumb_folder is None:
        output_thumb_folder = thumb_folder

    img, _ = models.ImageModel.get_or_create_from_path(img_path)
    def_thumb_rel, _ = models.ThumbnailRelationship.get_or_create_from_image(
        image=img,
        thumb_folder=output_thumb_folder,
        size=DEFAULT_SIZE,
        thumb_path=thumb_path,
        img_path=img_path
    )
    resized_thumb_rel = None

    if resize and size:
        resized_thumb_rel, _ = \
            models.ThumbnailRelationship.get_or_create_from_image(
                image=img,
                thumb_folder=output_thumb_folder,
                size=size,
                img_path=img_path
            )
    elif resize:
        # use thumbnail if no size is given
        resized_thumb_rel = def_thumb_rel
    else:
        # no resize, return models.ImageModel obj
        return img

    return resized_thumb_rel.thumbnail \
        if resized_thumb_rel is not None else img


def init_program(db_path=None):
    """Init program."""
    # create user data dir
    if not os.path.isdir(user_data_dir):
        os.makedirs(user_data_dir, exist_ok=True)
        log.debug('User data dir created.')
    # create thumbnail folder
    if not os.path.isdir(thumb_folder):
        os.makedirs(thumb_folder, exist_ok=True)
        log.debug('Thumbnail folder created.')

    # database
    if db_path is None:
        db_path = default_db_path
    models.init_db(db_path, db_version)


def get_tags_from_match_result(match_result, browser=None, scraper=None):
    """Get tags from match result."""
    if browser is None:
        browser = mechanicalsoup.StatefulBrowser(soup_config={'features': 'lxml'})
        browser.raise_on_404 = True

    res = models.MatchTagRelationship.select().where(
        models.MatchTagRelationship.match == match_result)
    tags = [x.tag for x in res]

    filtered_hosts = ['anime-pictures.net', 'www.theanimegallery.com']
    is_url_in_filtered_hosts = urlparse(match_result.link).netloc in \
        filtered_hosts
    if is_url_in_filtered_hosts:
        log.debug('URL in filtered hosts, no tag fetched', url=match_result.link)
    elif not tags:
        try:
            browser.open(match_result.link, timeout=10)
            page = browser.get_current_page()
            new_tags = get_tags_from_parser(page, match_result.link, scraper)
            new_tag_models = []
            if new_tags:
                for tag in new_tags:
                    namespace, tag_name = tag
                    tag_model, _ = models.Tag.get_or_create(
                        name=tag_name, namespace=namespace)
                    models.MatchTagRelationship.get_or_create(
                        match=match_result, tag=tag_model)
                    new_tag_models.append(tag_model)
            else:
                log.debug('No tags found.')

            tags.extend(new_tag_models)
        except (requests.exceptions.ConnectionError, mechanicalsoup.LinkNotFoundError) as e:
            log.error(str(e), url=match_result.link)
    return tags


def write_url_from_match_result(match_result, folder=None):
    """Write url from match result."""
    netloc = urlparse(match_result.link).netloc
    sanitized_netloc = netloc.replace('.', '_')
    text_file_basename = sanitized_netloc + '.txt'
    text_file = os.path.join(folder, text_file_basename) if folder is not None else text_file_basename
    with open(text_file, 'a') as f:
        f.write(match_result.link)
        f.write('\n')


def run_program_for_single_img(
        image, resize, size, place, match_filter, browser,
        scraper, disable_tag_print=False, write_tags=False, write_url=False
):
    """Run program for single image."""
    # compatibility
    br = browser

    error_set = []
    tag_textfile = image + '.txt'
    folder = os.path.dirname(image)
    result = []

    with NamedTemporaryFile() as temp, NamedTemporaryFile() as thumb_temp:
        shutil.copyfile(image, temp.name)
        post_img = get_posted_image(
            img_path=temp.name, resize=resize, size=size, thumb_path=thumb_temp.name)

        for img_m_rel_set in post_img.imagematchrelationship_set:
            for item_set in img_m_rel_set.imagematch_set:
                if item_set.search_place_verbose == place:
                    result.append(item_set)

        if not result:
            url, im_place = iqdb_url_dict[place]
            use_requests = True if place != 'e621' else False
            post_img_path = temp.name if not resize else thumb_temp.name
            page = get_page_result(
                image=post_img_path, url=url, browser=br,
                use_requests=use_requests
            )
            # if ok, will output: <Response [200]>
            result = list(models.ImageMatch.get_or_create_from_page(
                page=page, image=post_img, place=im_place))
            result = [x[0] for x in result]

    if match_filter == 'best-match':
        result = [x for x in result if x.status == x.STATUS_BEST_MATCH]

    log.debug('Number of valid result', n=len(result))
    for item in result:
        # type item: models.ImageMatch
        # type match_result: models.Match object
        match_result = item.match.match_result
        url = match_result.link
        print('{}|{}|{}'.format(
            item.similarity, item.status_verbose, url))

        try:
            tags = get_tags_from_match_result(match_result, browser, scraper)
            tags_verbose = [x.full_name for x in tags]
            log.debug('{} tag(s) founds'.format(len(tags_verbose)))
            if tags and not disable_tag_print:
                print('\n'.join(tags_verbose))

            if tags and write_tags:
                with open(tag_textfile, 'a') as f:
                    f.write('\n'.join(tags_verbose))
                    f.write('\n')
                log.debug('tags written')
            if write_url:
                write_url_from_match_result(match_result, folder)
        except Exception as e:  # pylint:disable=broad-except
            log.error('Error', e=str(e))
            error_set.append(e)

    return {'error': error_set}


@click.group()
def cli():
    """Run cli."""
    pass


@cli.command()
@click.option(
    '--place', type=click.Choice([x for x in iqdb_url_dict]),
    default=DEFAULT_PLACE,
    help='Specify iqdb place, default:{}'.format(DEFAULT_PLACE)
)
@click.option('--resize', is_flag=True, help='Use resized image.')
@click.option('--size', is_flag=True, help='Specify resized image.')
@click.option('--db-path', help='Specify Database path.')
@click.option(
    '--match-filter', type=click.Choice(['default', 'best-match']),
    default='default', help='Filter the result.'
)
@click.option('--write-tags', is_flag=True, help='Write best match\'s tags to text.')
@click.option('--write-url', is_flag=True, help='Write match url to text.')
@click.option(
    '--input-mode', type=click.Choice(['default', 'folder']),
    default='default', help='Set input mode.'
)
@click.option('--verbose', '-v', is_flag=True, help='Verbose output.')
@click.option('--debug', '-d', is_flag=True, help='Print debug output.')
@click.option(
    '--abort-on-error', is_flag=True, help='Stop program when error occured')
@click.argument('prog-input')
def run(
    prog_input=None, resize=False, size=None,
    db_path=None, place=DEFAULT_PLACE, match_filter='default',
    input_mode='default', verbose=False, debug=False,
    abort_on_error=False, write_tags=False, write_url=False

):
    """Get similar image from iqdb."""
    assert prog_input is not None, "Input is not a valid path"

    # logging
    log_level = None
    if verbose:
        log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    if log_level:

        logging.basicConfig(handlers=[logging.FileHandler(os.path.join(user_data_dir, 'output.log'), 'w', 'utf-8')], level=log_level)

    init_program(db_path)
    br = mechanicalsoup.StatefulBrowser(soup_config={'features': 'lxml'})
    br.raise_on_404 = True
    scraper = cfscrape.CloudflareScraper()

    # variable used in both input mode
    error_set = []
    if input_mode == 'folder':
        assert os.path.isdir(prog_input), 'Input is not valid folder'
        files = [os.path.join(prog_input, x) for x in os.listdir(prog_input)]
        if not files:
            print('No files found.')
            return
        sorted_files = sorted(files, key=lambda x: os.path.splitext(x)[1])
        for idx, ff in enumerate(sorted_files):
            log.debug(
                'file', f=os.path.basename(ff), idx=idx, total=len(files))
            result = {}
            try:
                result = run_program_for_single_img(
                    ff, resize, size, place, match_filter,
                    browser=br, scraper=scraper, disable_tag_print=True,
                    write_tags=write_tags, write_url=write_url
                )
            except Exception as e:  # pylint:disable=broad-except
                if abort_on_error:
                    raise e
                error_set.append((ff, e))
            if result is not None and result.get('error'):
                error_set.extend([(ff, x) for x in result['error']])
    else:
        image = prog_input
        result = run_program_for_single_img(
            image, resize, size, place, match_filter,
            browser=br, scraper=scraper,
            write_tags=write_tags, write_url=write_url
        )
        if result is not None and result.get('error'):
            error_set.extend([(image, x) for x in result['error']])

    if error_set:
        log.error('Found error(s)')
        for x in error_set:
            log.error('path: ' + x[0] + '\nerror: ' + str(x[1]))


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
"""Utils module."""
import os

from appdirs import user_data_dir

user_data_dir = user_data_dir('iqdb_tagger', 'softashell')
default_db_path = os.path.join(user_data_dir, 'iqdb.db')
thumb_folder = os.path.join(user_data_dir, 'thumbs')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""server module."""
from logging.handlers import TimedRotatingFileHandler
import logging
import os
import pprint

from flask import (
    Flask,
    send_from_directory,
)
from flask.cli import FlaskGroup
from flask_admin import Admin
# from flask_admin.contrib.peewee import ModelView
import click
import structlog

from iqdb_tagger.__main__ import init_program
from iqdb_tagger.models import init_db
from iqdb_tagger.utils import default_db_path, thumb_folder, user_data_dir
from iqdb_tagger import views

log = structlog.getLogger()


def thumb(basename):
    """Get thumbnail."""
    return send_from_directory(thumb_folder, basename)


def create_app(script_info=None):
    """Create app."""
    app = Flask(__name__)
    # logging
    if not os.path.exists(user_data_dir):
        os.makedirs(user_data_dir)
    log_dir = os.path.join(user_data_dir, 'log')
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    peewee_logger = logging.getLogger('peewee')
    peewee_logger.setLevel(logging.INFO)
    chardet_logger = logging.getLogger('chardet')
    chardet_logger.setLevel(logging.INFO)
    default_log_file = os.path.join(log_dir, 'iqdb_tagger_server.log')
    file_handler = TimedRotatingFileHandler(default_log_file, 'midnight')
    file_handler.setLevel(logging.WARNING)
    file_handler.setFormatter(logging.Formatter('<%(asctime)s> <%(levelname)s> %(message)s'))
    app.logger.addHandler(file_handler)
    app.logger.addHandler(peewee_logger)
    app.logger.addHandler(chardet_logger)
    # reloader
    reloader = app.config['TEMPLATES_AUTO_RELOAD'] = \
        bool(os.getenv('IQDB_TAGGER_RELOADER')) or app.config['TEMPLATES_AUTO_RELOAD']  # NOQA
    if reloader:
        app.jinja_env.auto_reload = True
    app.config['SECRET_KEY'] = os.getenv('IQDB_TAGGER_SECRET_KEY') or os.urandom(24)
    app.config['WTF_CSRF_ENABLED'] = False
    # debug
    debug = app.config['DEBUG'] = bool(os.getenv('IQDB_TAGGER_DEBUG')) or app.config['DEBUG']
    if debug:
        app.config['DEBUG'] = True
        app.config['LOGGER_HANDLER_POLICY'] = 'debug'
        logging.basicConfig(level=logging.DEBUG)
        pprint.pprint(app.config)
        print('Log file: {}'.format(default_log_file))
        print('script info:{}'.format(script_info))
    db_path = os.getenv('IQDB_TAGGER_DB_PATH') or default_db_path
    init_program()
    init_db(db_path)
    # app and db
    app.app_context().push()

    @app.shell_context_processor
    def shell_context():  # pylint: disable=unused-variable
        return {'app': app}

    # flask-admin
    app_admin = Admin(
        app, name='IQDB Tagger', template_mode='bootstrap3',
        index_view=views.HomeView(name='Home', template='iqdb_tagger/index.html', url='/'))

    app_admin.add_view(views.MatchView())
    # app_admin.add_view(ModelView(ImageMatch, category='DB'))
    # app_admin.add_view(ModelView(ImageMatchRelationship, category='DB'))
    # app_admin.add_view(ModelView(ImageModel, category='DB'))
    # app_admin.add_view(ModelView(MatchTagRelationship, category='DB'))
    # routing
    app.add_url_rule('/thumb/<path:basename>', view_func=thumb)
    return app


@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    """Run cli. This is a management script for application."""
    pass


if __name__ == '__main__':
    cli()



"""forms module."""
# pylint: disable=ungrouped-imports
from flask_wtf import FlaskForm
from wtforms import BooleanField, SelectField
from flask_wtf.file import FileField, FileAllowed, FileRequired

from . import models


class ImageUploadForm(FlaskForm):
    """Image upload form."""

    file = FileField('Image', validators=[
        FileRequired(),
        FileAllowed(['jpg', 'png'], 'Images only!')
    ])
    resize = BooleanField()
    place = SelectField('Place', choices=models.ImageMatch.SP_CHOICES)



"""views module."""
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse

from flask import request, redirect, url_for, current_app, flash, abort
from flask_admin import AdminIndexView, expose, BaseView
from flask_paginate import Pagination, get_page_parameter
import requests

from .models import ImageModel, ImageMatchRelationship, ImageMatch
from . import forms, models
from .__main__ import (
    get_page_result,
    get_posted_image,
    get_tags_from_match_result,
    iqdb_url_dict,
)


class HomeView(AdminIndexView):
    """Home view."""

    @expose('/', methods=('GET', 'POST'))
    def index(self):
        """Get index page."""
        form = forms.ImageUploadForm()
        if form.file.data:
            print('resize:{}'.format(form.resize.data))
            with NamedTemporaryFile() as temp, NamedTemporaryFile() as thumb_temp:
                form.file.data.save(temp.name)
                posted_img = get_posted_image(
                    img_path=temp.name, resize=form.resize.data, thumb_path=thumb_temp.name)
                place = [x[1] for x in form.place.choices if x[0] == int(form.place.data)][0]
                url, im_place = iqdb_url_dict[place]
                query = posted_img.imagematchrelationship_set \
                    .select().join(ImageMatch) \
                    .where(ImageMatch.search_place == im_place)
                if not query.exists():
                    try:
                        posted_img_path = temp.name if not form.resize.data else thumb_temp.name
                        result_page = get_page_result(image=posted_img_path, url=url)
                    except requests.exceptions.ConnectionError as e:
                        current_app.logger.error(str(e))
                        flash('Connection error.')
                        return redirect(request.url)
                    list(ImageMatch.get_or_create_from_page(
                        page=result_page, image=posted_img, place=im_place))
            return redirect(url_for('matchview.match_sha256', checksum=posted_img.checksum))

        page = request.args.get(get_page_parameter(), type=int, default=1)
        item_per_page = 10
        entries = (
            ImageModel.select()
            .distinct()
            .join(ImageMatchRelationship)
            .where(ImageMatchRelationship.image)
            .order_by(ImageModel.id.desc())
        )
        pagination = Pagination(
            page=page, total=entries.count(), per_page=item_per_page, bs_version=3)
        paginated_entries = entries.paginate(page, item_per_page)
        if not entries.exists() and page != 1:
            abort(404)
        # pagination = Pagination(page, item_per_page, entries.count())
        return self.render(
            'iqdb_tagger/index.html',
            entries=paginated_entries,
            pagination=pagination,
            form=form
        )


class MatchView(BaseView):
    """Match view."""

    @expose('/')
    def index(self):
        """Index page."""
        return self.render('iqdb_tagger/match.html')

    @expose('/sha256-<checksum>')
    def match_sha256(self, checksum):
        """Get image match the checksum."""
        current_app.logger.debug('match sha256: {}'.format(request.url))
        entry = models.ImageModel.get(models.ImageModel.checksum == checksum)
        return self.render('iqdb_tagger/match_checksum.html', entry=entry)

    @expose('/d/<pair_id>')
    def match_detail(self, pair_id):
        """Show single match pair."""
        nocache = False
        entry = ImageMatchRelationship.get(ImageMatchRelationship.id == pair_id)

        match_result = entry.match_result
        mt_rel = models.MatchTagRelationship.select().where(
            models.MatchTagRelationship.match == match_result)
        tags = [x.tag.full_name for x in mt_rel]
        filtered_hosts = ['anime-pictures.net', 'www.theanimegallery.com']

        if urlparse(match_result.link).netloc in filtered_hosts:
            current_app.logger.debug(
                'URL in filtered hosts, no tag fetched, url:{}'.format(match_result.link))
        elif not tags or nocache:
            try:
                tags = list(get_tags_from_match_result(match_result))
                if not tags:
                    current_app.logger.debug('Tags not founds, id:{}'.format(pair_id))
            except requests.exceptions.ConnectionError as e:
                current_app.logger.debug(str(e) + 'url:{}'.format(match_result.link))
        return self.render('iqdb_tagger/match_single.html', entry=entry)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""sha256 module."""
import hashlib
import sys


def sha256_checksum(filename, block_size=65536):
    """Get sha256 checksum."""
    sha256 = hashlib.sha256()
    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()


def main():
    """Run main func for module."""
    for f in sys.argv[1:]:
        checksum = sha256_checksum(f)
        print(f + '\t' + checksum)


if __name__ == '__main__':
    main()



"""parser module."""
import bs4
import cfscrape
import structlog

log = structlog.getLogger()


def get_tags(page, url, scraper=None):
    """Get tags by parsing page from the url.

    url is used as hint to choose which parser will be used.
    """
    parser = [
        ChanSankakuParser,
        DanbooruParser,
        E621Parser,
        Eshuushuu,
        GelbooruParser,
        Konachan,
        YandereParser,
        ZerochanParser,
    ]
    log.debug('url', url=url)
    for item in parser:
        obj = item(url, page, scraper)
        if obj.is_url(url):
            log.debug('match', parser=item)
            result = list(obj.get_tags())
            return result
    log.debug('No parser found', url=url)


class CustomParser:
    """Base for custom parser."""

    def __init__(self, url, page, scraper=None):
        """Init method."""
        self.url = url
        self.page = page
        self.scraper = scraper

    @staticmethod
    def is_url(url):
        """Check url."""
        raise NotImplementedError

    def get_tags(self):
        """Get tags."""
        raise NotImplementedError


class YandereParser(CustomParser):
    """Parser for yande.re."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'yande.re/post/show/' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        page = self.page
        classname_to_namespace_dict = {
            'tag-type-copyright': 'series',
            'tag-type-character': 'character',
            'tag-type-general': ''
        }
        for key, value in classname_to_namespace_dict.items():
            for item in page.select('li.{}'.format(key)):
                text = item.text.strip().split(' ', 1)[1].rsplit(' ', 1)[0]
                yield (value, text)


class ChanSankakuParser(CustomParser):
    """Parser for chan.sankakucomplex."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'chan.sankakucomplex.com/post/show' in url:
            return True
        return False

    @staticmethod
    def parse_page(page):
        """Parse page."""
        classname_to_namespace_dict = {
            'tag-type-artist': 'creator',
            'tag-type-character': 'character',
            'tag-type-copyright': 'series',
            'tag-type-meta': 'meta',
            'tag-type-general': ''
        }
        for key, namespace in classname_to_namespace_dict.items():
            for item in page.select('li.{}'.format(key)):
                name = item.text.rsplit('(?)', 1)[0].strip()
                yield (namespace, name)

    def get_tags(self):
        """Get tags."""
        page = self.page
        url = self.url
        result = list(self.parse_page(page))
        if not result:
            h1_tag_text = page.select_one('h1').text
            if h1_tag_text != '503 Service Temporarily Unavailable':
                log.error('Unexpected H1-tag text', text=h1_tag_text)
            if self.scraper is None:
                self.scraper = cfscrape.CloudflareScraper()
            resp = self.scraper.get(url, timeout=10)
            html_soup = bs4.BeautifulSoup(resp.text, 'lxml')
            return self.parse_page(html_soup)
        return result


class GelbooruParser(CustomParser):
    """Parser for gelbooru.com."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'gelbooru.com/index.php?' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        page_title = self.page.select_one('head title').text.strip()
        if page_title == 'Image List  | Gelbooru':
            log.debug('Image list instead of post found.', url=self.url)
            return
        page = self.page
        classname_to_namespace_dict = {
            'tag-type-artist': 'creator',
            'tag-type-character': 'character',
            'tag-type-copyright': 'series',
            'tag-type-general': ''
        }
        for key, value in classname_to_namespace_dict.items():
            for item in page.select('li.{}'.format(key)):
                try:
                    text = item.text.rsplit(' ', 1)[0].split(' ', 1)[1].strip()
                except IndexError:
                    new_item_text = item.text.replace('\n', ' ')
                    new_item_text = new_item_text.rsplit(' ', 1)[0].strip()
                    new_item_text = new_item_text.split('? + - ', 1)[1]
                    text = new_item_text
                yield(value, text)


class ZerochanParser(CustomParser):
    """Parser for zerochan."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'www.zerochan.net/' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        page = self.page
        tags = page.select('ul#tags li')
        for tag in tags:
            try:
                tag_text, namespace = tag.text.rsplit(' ', 1)
                yield (namespace, tag_text)
            except TypeError as e:
                log.error(str(e), tag_text=tag)
                yield ('', '')


class DanbooruParser(CustomParser):
    """Parser for danbooru."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'danbooru.donmai.us/posts/' in url:

            return True
        return False

    def get_tags(self):
        """Get tags."""
        page = self.page
        classname_to_namespace_dict = {
            'category-0': '',
            'category-1': 'creator',
            'category-2': 'meta',
            'category-3': 'series',
            'category-4': 'character',
            'category-5': 'meta',
            'category-6': 'meta',
            'category-7': 'meta',
        }
        for key, value in classname_to_namespace_dict.items():
            for item in page.select('li.{}'.format(key)):
                text = item.text.rsplit(' ', 1)[0].split(' ', 1)[1].strip()
                yield value, text


class Eshuushuu(CustomParser):
    """Parser for e-shuushuu.net."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'e-shuushuu.net/image/' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        page = self.page
        classname_to_namespace_dict = {
            'quicktag1_': '',
            'quicktag2_': 'series',
            'quicktag3_': 'creator',
            'quicktag4_': 'character',
        }
        for classname, namespace in classname_to_namespace_dict.items():
            tags = page.select(
                'div.meta dd[id^={}] span.tag a'.format(classname))
            for tag in tags:
                yield (namespace, tag.text)


class Konachan(CustomParser):
    """Parser for konachan.com."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'konachan.com/post/show/' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        page = self.page
        classname_to_namespace_dict = {
            'tag-type-artist': 'creator',
            'tag-type-character': 'character',
            'tag-type-circle': 'character',
            'tag-type-copyright': 'series',
            'tag-type-style': 'style',
            'tag-type-general': ''
        }
        for classname, namespace in classname_to_namespace_dict.items():
            tags = page.select('ul li.{}'.format(classname))
            for tag in tags:
                text = tag.text.split(' ', 1)[1].strip().rsplit(' ', 1)[0]
                yield namespace, text


class E621Parser(CustomParser):
    """Parser for e621."""

    @staticmethod
    def is_url(url):
        """Check url."""
        if 'e621.net/post/show/' in url:
            return True
        return False

    def get_tags(self):
        """Get tags."""
        classname_to_namespace_dict = {
            'tag-type-artist': 'creator',
            'tag-type-character': 'character',
            'tag-type-copyright': 'series',
            'tag-type-species': 'species',
            'tag-type-general': ''
        }
        scraper = cfscrape.CloudflareScraper()
        resp = scraper.get(self.url, timeout=10)
        page = bs4.BeautifulSoup(resp.text, 'lxml')

        for key, namespace in classname_to_namespace_dict.items():
            for item in page.select('li.{}'.format(key)):
                name = \
                    item.text \
                    .rsplit(' ', 1)[0].strip().split('? ', 1)[1].strip()
                yield (namespace, name)



"""setup file."""
from os import path

from setuptools import find_packages, setup

__version__ = '0.2.0'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

install_requires = [
    'appdirs>=1.4.3',
    'beautifulsoup4>=4.6.0',
    'cfscrape==1.9.0',
    'click>=6.7',
    'Flask-Admin==1.5.1',
    'flask-paginate==0.5.1',
    'Flask-WTF==0.14.2',
    'Flask==0.12.2',
    'Flask>=0.12.2',
    'funclog>=0.3.0',
    'lxml>=3.8.0',
    'MechanicalSoup>=0.7.0',
    'peewee>=2.10.1',
    'Pillow>=4.2.1',
    'structlog>=17.2.0',
    'wtf-peewee>=3.0.0',
]

setup(
    name='iqdb_tagger',
    version=__version__,
    description='Get result from iqdb.org.',
    long_description=long_description,
    url='https://github.com/softashell/iqdb_tagger',
    download_url='https://github.com/softashell/iqdb_tagger'
    '/tarball/' + __version__,
    license='BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],
    keywords='',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='softashell',
    author_email='foreturiga@gmail.com',
    maintainer='Rachmadani Haryono',
    maintainer_email='foreturiga@gmail.com',
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'iqdb-tagger = iqdb_tagger.__main__:cli',
            'iqdb-tagger-server = iqdb_tagger.server:cli',
        ]
    }

)
